/* SPDX-License-Identifier: GPL-2.0 */
/*
 * main.c - for periodicstats everything starts here.
 *
 * Copyright (C) 2021 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Shubham Sharma <shubham.sharma@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>

#include <libubus.h>

#include "periodicstats.h"

// Global base interval
int base_interval = 0;

// Global periodic stats structure
static LIST_HEAD(ss_list);

struct ubus_context *ctx = NULL;

/**
 *  This function prints the usage of the daemon
 *  @param prog input pointer to char, for program name
 */
static void pstat_usage(char *prog)
{
	fprintf(stderr, "Usage: %s [options]\n", prog);
	fprintf(stderr, "\n");
	fprintf(stderr, "options:\n");
	fprintf(stderr, "    -s <socket path>   ubus socket\n");
	fprintf(stderr, "    -h <help>       To print this help menu\n");
	fprintf(stderr, "\n");
}

/**
 *  fetch_base_interval function to fetch globally configured base interval
 *  retrun poistive integer if configured and -1 on failure
 */
static int fetch_base_interval(void)
{
	int ret = 0;
	char *opn = NULL;
	struct uci_package *uci_pkg = NULL;
	struct uci_context *uci_ctx = uci_alloc_context();

	uci_load(uci_ctx, "periodicstats", &uci_pkg);
	if (!uci_pkg) {
		syslog(LOG_ERR, "Failed to load configuration\n");
		ret = -1;
		uci_free_context(uci_ctx);
		return ret;
	}

	struct uci_section *sec = uci_lookup_section(uci_ctx, uci_pkg, "globals");
	if (sec == NULL) {
		syslog(LOG_ERR, "Unable to find globals uci section\n");
		ret = -1;
		goto done;
	}

	// Global enable/diable for periodicstats
	opn = (char *)uci_lookup_option_string(uci_ctx, sec, "base_interval");
	if (opn == NULL) {
		syslog(LOG_ERR, "Unable to find globals uci enable option\n");
		ret = -1;
		goto done;
	}

	if (!atoi(opn)) {
		ret = -1;
		goto done;
	}

	ret = atoi(opn);

done:
	uci_unload(uci_ctx, uci_pkg);
	uci_free_context(uci_ctx);
	return ret;
}

/**
 *  Main function for periodicstats, everything starts here
 *  @param argc input number of input arguments
 *  @param argv input double pointer array of optional command line arguments
 *  retrun integer value 0 on success and -1 on failure
 */
int main(int argc, char **argv)
{
	int ch;
	int ret;
	const char *ubus_socket = NULL;

	while ((ch = getopt(argc, argv, "hs:")) != -1) {
		switch (ch) {
		case 's':
			ubus_socket = optarg;
			break;
		case 'h':
			pstat_usage(argv[0]);
			exit(0);
		default:
			break;
		}
	}

	/* Logging to syslog */
	openlog("periodicstats", LOG_PID|LOG_CONS, LOG_LOCAL1);

	/* stats container directory init */
	mkdir(SAMPLE_SET_DIR, 0755);

	argc -= optind;
	argv += optind;

	uloop_init();
	ctx = ubus_connect(ubus_socket);
	if (!ctx) {
		syslog(LOG_ERR, "Failed to connect to ubus\n");
		return -1;
	}

	// Fetch global base_interval
	base_interval = fetch_base_interval();

	// Initialize global periodic stats structure
	ret = pstats_global_init(&ss_list);
	if (ret) {
		syslog(LOG_ERR, "Failed to initialize global periodic stats.\n");
		goto out;
	}

	ubus_add_uloop(ctx);

	// stats collection invoked from here
	pstats_get_stats(&ss_list);

	/* Main loop of periodicstats */
	uloop_run();
	pthread_exit(NULL);

 out:
	ubus_free(ctx);
	uloop_done();

	return 0;
}
