/* SPDX-License-Identifier: GPL-2.0 */
/*
 * periodicstats.c - periodic stats method handeling code
 *
 * Copyright (C) 2021 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Shubham Sharma <shubham.sharma@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <time.h>

#include <libubus.h>
#include <libubox/blobmsg_json.h>

#include <libbbf_api/dmcommon.h>

#include "periodicstats.h"

#define BASE 10

typedef struct thread_arguments {
	smpl_set *pos;
	param *ptr;
	char sampleset_dir[MAX_REF_STR];
} thrd_args;

pthread_mutex_t lock;

/**
 *  pstats_write_to_file function to final stats writing to specified parameter per sampleset file
 *  @param result input pointer to json_object, containing final stats result
 *  @param ptr input pointer to param structure
 *  @param sampleset_dir input pointer to char, containing file name in which result to be written
 */
static void pstats_write_to_file(json_object *result, param *ptr, char *sampleset_dir)
{
	FILE *fptr;
	char file_name[MAX_NAME_STR] = {0};

	snprintf(file_name, sizeof(file_name), "%s%s", sampleset_dir, ptr->name);
	fptr = fopen(file_name, "w");
	if (fptr == NULL) {
		syslog(LOG_ERR, "fopen() Error on : %s\n", file_name);
	}

	fprintf(fptr, "%s\n", json_object_to_json_string(result));

	fclose(fptr);
}

static void pstats_remove_first_data(char *prev_val, char *new_val, size_t length)
{
	size_t len;

	if (prev_val == NULL) {
		return;
	}
	char *temp = strtok(prev_val, ",");

	len = strlen(temp);
	strncpy(new_val, prev_val+len+1, length);
}

static void pstats_update_suspect_data(json_object *result, unsigned int update)
{
	json_object *val;
	char new_val[MAX_TIME_STR] = {'\0'};
	char updated_data[MAX_DATA_STR] = {'\0'};

	json_object_object_get_ex(result, "SuspectData", &val);

	pstats_remove_first_data((char *)json_object_get_string(val), updated_data, sizeof(updated_data));

	snprintf(new_val, sizeof(new_val), "%u,", update);
	strncat(updated_data, new_val, sizeof(new_val) - strlen(new_val));

	json_object_object_add(result,"SuspectData", json_object_new_string(updated_data));
}

/**
 *  pstats_process_calculation_mode function to handle Calculation Mode
 *  @param t_args input void pointer to thrd_args, containing sampleset and parameter uci information
 *  @param data input integer value
 *  @param prev input integer value
 *  @param final output pointer to character string
 */
void pstats_process_calculation_mode(thrd_args *t_args, int data, int *prev, int avg_counter, char *final)
{
	switch(t_args->ptr->param_cal_mode) {

		case LATEST:
			sprintf(final, "%d", data);
			break;

		case MINIMUM:
			if (data < *prev) {
				sprintf(final, "%d", data);
			} else {
				sprintf(final, "%d", *prev);
			}
			break;

		case MAXIMUM:
			if (data > *prev) {
				sprintf(final, "%d", data);
			} else {
				sprintf(final, "%d", *prev);
			}
			break;

		case AVERAGE:
			if (avg_counter > 0) {
				*prev += data;
			} else {
				int samples = t_args->pos->smpl_intrvl / base_interval;
				*prev = (*prev + data) / samples;
			}

			sprintf(final, "%d", *prev);
			break;
	}
}

/**
 *  pstats_calculation_mode_collection function to read from database, this function is also
 *  responsible for updating the value in case its an interger and mode is other than
 *  latest (min, max or average). It fills the value finally to be displayed at tr181
 *  @param t_args input pointer to thrd_args structure containing complete configuration data
 *  @param value output pointer to char containg the final read value
 *  @param susp_data_flag output pointer to susp_data_flag
 */
static void pstats_calculation_mode_collection(thrd_args *t_args, char *value, bool *susp_data_flag)
{
	int prev = 0;
	int tm_counter = 0;
	char *endptr = NULL;

	bool first_value = true;

	int avg_counter = t_args->pos->smpl_intrvl / base_interval;

	struct ubus_arg arg = {"path", t_args->ptr->reference, String};

	while(tm_counter < t_args->pos->smpl_intrvl) {
		json_object *jobj= NULL;
		enum json_type type;

		pthread_mutex_lock(&lock);
		dmubus_call("usp", "get", &arg, 1, &jobj);
		if (jobj == NULL) {
			syslog(LOG_ERR, "Could not read stats");
			break;
		}

		json_object_object_foreach(jobj, key, val) {
			BBF_ATTR_UNUSED(key);
			type = json_object_get_type(val);
			switch(type) {
			case json_type_int:
				if (first_value) {
					prev = json_object_get_int(val);
					strncpy(value, json_object_get_string(val),strlen(value));
					first_value = false;
				} else {
					pstats_process_calculation_mode(t_args, json_object_get_int(val), &prev, avg_counter, value);
					prev = strtol(value, &endptr, BASE);
					avg_counter--;
				}
				break;
			default:
				strncpy(value, json_object_get_string(val),strlen(value));
				break;

			}
			break;
		}
		dmubus_free();
		pthread_mutex_unlock(&lock);

		tm_counter += base_interval;
		sleep(base_interval);
	}
}

/**
 *  pstats_refresh_stats function to periodically update statistics
 *  @param fname input string name of file to be read from
 *  @param slit input pointer to start time for stats collection to be updated
 *  @param t_args input pointer to thread arguments which are nothing but the config
 *                      values according to which the stats collection is to be performed
 */
static int pstats_refresh_stats(char *fname, struct tm *slt, thrd_args *t_args)
{
	int i, len;
	bool susp_data_flag = true;
	char start_tm[MAX_TIME_STR] = {'\0'};

	json_object *str = NULL;
	char value[MAX_DATA_STR];

	// get the latest set of values
	pstats_calculation_mode_collection(t_args, value, &susp_data_flag);

	// start updating the json object holding results
	json_object *reference, *values;
	json_object *result = json_object_from_file(fname);

	if (result == NULL) {
		// something is really wrong, this means that the report
		// file is not even existing, no point continuing
		syslog(LOG_ERR, "report file not created");
		return -1;
	}

	// extract reference from the json read from file
	json_object_object_get_ex(result, "Values", &values);
	json_object_object_get_ex(values, t_args->ptr->reference, &reference);

	strftime(start_tm, sizeof(start_tm), "%Y-%m-%dT%H:%M:%SZ", slt);
	json_object *start_time = json_object_new_string(start_tm);
	json_object_object_add(result,"ReportStartTime", start_time);

	len = json_object_array_length(reference);

	// Shift values left in the array to accomodate stacking of latest value
	for (i = 0; i < (len-1); i++) {
		json_object *pos;

		pos = json_object_array_get_idx(reference, i+1);
		json_object_array_put_idx(reference, i, json_object_get(pos));
	}

	// update latest value to end of string
	str = json_object_new_string(value);
	json_object_array_put_idx(reference, (len-1), str);

	if (susp_data_flag)
		pstats_update_suspect_data(result, 1);
	else
		pstats_update_suspect_data(result, 0);

	// End of data collection
	time_t et;
	time(&et);

	char end_tm[MAX_TIME_STR] = {'\0'};
	struct tm *elt = localtime(&et);
	strftime(end_tm, sizeof(end_tm), "%Y-%m-%dT%H:%M:%SZ", elt);

	json_object *end_time = json_object_new_string(end_tm);
	json_object_object_add(result,"ReportEndTime", end_time);

	pthread_mutex_lock(&lock);
	pstats_write_to_file(result, t_args->ptr, t_args->sampleset_dir);
	pthread_mutex_unlock(&lock);
	json_object_put(result);

	return 0;
}

/**
 *  pstats_loopover_latest_samples function to update latest samples to
 *  collected samples
 *  @param st input to time_t, report start time reference to be updated
 *  @param t_args input pointer to thrd_args structure containing uci config info
 *                according to which stats are to be collected
 */
static void pstats_loopover_latest_samples(time_t *st, thrd_args *t_args)
{
	struct tm *slt = localtime(st);
	char *fname;
	size_t len = sizeof(t_args->sampleset_dir)+ sizeof( t_args->ptr->name);

	fname = (char *)calloc(len, sizeof(char));
	if(fname == NULL) {
		syslog(LOG_ERR, "calloc failed.\n");
		return ;
	}
	snprintf(fname,len, "%s%s", t_args->sampleset_dir, t_args->ptr->name);

	while (1) {
		// update Report Start time
		*st += t_args->pos->smpl_intrvl;
		slt = localtime(st);

		// Now get refresh the values
		if (0 != pstats_refresh_stats(fname, slt, t_args))
			break;
	}
	free(fname);
}

/**
 *  pstats_get_actual_start_time function to determine actual starts time based on time_reference configured,
 *  this function is also responsible to wait till its time t begin sample collection
 *  @param args input/output pointer to start time
 *  @param args input time_t for reference time
 *  @param args input integer for sample interval
 */
static void pstats_get_actual_start_time(time_t *st, time_t time_ref, int smpl_intrvl)
{
	time_t required_delay;
	char launch[MAX_REF_STR] = {0};

	time_t ref_delay = (time_t)difftime(time_ref, *st);

	if (ref_delay >= 0) {
		// Filling required delay with just the difference
		required_delay = ref_delay;

	} else {
		// tr181 defines the time ref as the time at which the first sample is collected,
		// so we need to subtract the sample interval from time-ref to get the start time
		// of sample collections so that the first sample is made available at time_ref
		required_delay = smpl_intrvl - (abs(ref_delay) % smpl_intrvl);
	}

	*st += required_delay;

	strftime(launch, sizeof(launch), "%Y-%m-%dT%H:%M:%SZ", localtime(st));

	// Wait until actual start time reach
	sleep(required_delay);
}

/**
 *  pstats_samples function to get corresponding value of the referenced parameter. This function
 *  is used for the first time population of the values.
 *  @param t_args input void pointer to thrd_args, containing uci values for sampleset and parameter
 *  @param suspect_data pointer to comma seperated list of suspected data flag values
 *  @param sample_seconds pointer to comma seperated list of elapsed seconds from start at which
 *                        stats were collected
 *  @param start_tm pointer to time at which stat collection was started
 *  @param seconds input unsigned integer value
 */
static void pstats_samples(thrd_args *t_args, char *suspect_data,
		char *sample_seconds, char *start_tm, unsigned int seconds)
{
	time_t et;
	struct tm *elt;
	bool susp_data_flag = true;
	char end_tm[MAX_TIME_STR] = {'\0'};

	char smpl_sec[MAX_TIME_STR] = {'\0'};
	char susp_data[MAX_TIME_STR] = {'\0'};

	json_object *str = NULL;
	char value[MAX_REF_STR];
	
	// stats collection based on calculation mode
	pstats_calculation_mode_collection(t_args, value, &susp_data_flag);

	char *fname ;
	size_t len = sizeof(t_args->sampleset_dir)+ sizeof( t_args->ptr->name);
	
	fname = (char *) calloc(len, sizeof(char));
	if(fname == NULL) {
		syslog(LOG_ERR, "calloc failed.\n");
		return ;
	}
	snprintf(fname, len, "%s%s", t_args->sampleset_dir, t_args->ptr->name);

	json_object *reference, *values;
	json_object *result;

	if (seconds == 0) {
		json_object *start_time;
		// first sample collection, create object
		result = json_object_new_object();
		values = json_object_new_object();
		reference = json_object_new_array();
		start_time = json_object_new_string(start_tm);
		json_object_object_add(result,"ReportStartTime", start_time);
		json_object_object_add(values, t_args->ptr->reference, reference);
		json_object_object_add(result,"Values", values);
	} else {
		result = json_object_from_file(fname);
		// extract reference from the json read from file
		json_object_object_get_ex(result, "Values", &values);
		json_object_object_get_ex(values, t_args->ptr->reference, &reference);

	}
	str = json_object_new_string(value);
	json_object_array_add(reference, str);

	if (susp_data_flag)
		snprintf(susp_data, sizeof(susp_data), "%d,", 1);
	else
		snprintf(susp_data, sizeof(susp_data), "%d,", 0);

	strncat(suspect_data, susp_data, MAX_DATA_STR - 1 - strlen(suspect_data));
	snprintf(smpl_sec, sizeof(smpl_sec), "%u,", seconds);

	strncat(sample_seconds, smpl_sec, MAX_DATA_STR - 1 - strlen(sample_seconds));

	json_object *sample_sec = json_object_new_string(sample_seconds);
	json_object_object_add(result,"SuspectData", json_object_new_string(suspect_data));
	json_object_object_add(result,"SampleSeconds", sample_sec);

	time(&et);
	elt = localtime(&et);
	strftime(end_tm, sizeof(end_tm), "%Y-%m-%dT%H:%M:%SZ", elt);

	json_object *end_time = json_object_new_string(end_tm);
	json_object_object_add(result,"ReportEndTime", end_time);

	pthread_mutex_lock(&lock);
	pstats_write_to_file(result, t_args->ptr, t_args->sampleset_dir);
	pthread_mutex_unlock(&lock);

	json_object_put(result);

	free(fname);
}

/**
 *  pstats_collection pthread handler function, actual periodicstattistics collection starts here
 *  @param args input void pointer to thrd_args, containing sampleset and parameter uci information
 *                    according to which the sample collection is done.
 */
static void *pstats_collection(void *args)
{
	time_t st;
	unsigned int seconds = 0;
	char start_tm[MAX_TIME_STR] = {'\0'};
	char suspect_data[MAX_DATA_STR] = {'\0'};
	char sample_seconds[MAX_DATA_STR] = {'\0'};
	thrd_args *t_args = (thrd_args *)args;

	time(&st);

	// Decide whether to start the collection right away or do we need to wait,
	// the function pstats_get_actual_start_time sleeps by itself if wait is needed.
	if (t_args->pos->time_ref)
		pstats_get_actual_start_time(&st, t_args->pos->time_ref, t_args->pos->smpl_intrvl);

	strftime(start_tm, sizeof(start_tm), "%Y-%m-%dT%H:%M:%SZ", localtime(&st));

	// This is the case when the SampleSet.Parameter is enabled for the first
	// time and no value exists for this object
	for (int i = 0; i < t_args->pos->rprt_smpls; i++) {
		pstats_samples(t_args, suspect_data, sample_seconds, start_tm, seconds);
		seconds += t_args->pos->smpl_intrvl;
	}

	// This is the handling for the n+1 case, for example, if the value of the
	// number of reported samples is 10, then, after the collection of initial
	// 10 values, we need to stack the stats, that is, the first stat value
	// should be lost from bottom and the latest should be put on top
	pstats_loopover_latest_samples(&st, t_args);

	free(t_args->ptr);
	free(t_args->pos);
	free(t_args);
	return NULL;
}

/**
 *  pstats_launch_thread function to spawn thread per sampleset parameter
 *  @param args input pointer to thrd_args, containing sampleset and parameter uci information
 */
void pstats_launch_thread(thrd_args *args)
{
	int res;
	pthread_t tid;

	res = pthread_create(&tid, NULL, pstats_collection, (void *)args);
	if ( res != 0 ) {
		syslog(LOG_ERR, "pthread_create() failed.\n");
		return;
	} else {
		res = pthread_detach(tid);
		if ( res != 0 ) {
			syslog(LOG_ERR, "pthread_detach() failed.\n");
			return;
		}
	}
}

/**
 *  pstats_get_stats function for invoking periodic statistics
 *  @param global_list input pointer to global list_head structure
 */
void pstats_get_stats(struct list_head *global_list)
{
	if (pthread_mutex_init(&lock, NULL) != 0) {
		syslog(LOG_ERR, "Mutex initialization failed.\n");
		return;
	}

	smpl_set *pos = NULL; // pointer to structure that carries the uci section sample_set params
	list_for_each_entry(pos, global_list, list) {
		if (pos->enable == 1) {
			char sampleset_dir[MAX_REF_STR] = {0};
			snprintf(sampleset_dir, sizeof(sampleset_dir), "%s%s/", SAMPLE_SET_DIR, pos->name);
			mkdir(sampleset_dir, 0755);

			param *ptr = NULL; // pointer to uci section parameter mapping to SampeSet.Parameter
			list_for_each_entry(ptr, &pos->params, list) {
				if ((ptr->enable == 1) && (ptr->reference[0] != '\0')) {
					// For each parameter instance, launch a seprate thread to
					// collect and store stats

					thrd_args *args ;

					args = (thrd_args *)calloc(1, sizeof(thrd_args));
					if (args == NULL)
						return;
					args->pos = (smpl_set *)calloc(1, sizeof(smpl_set));
					if (args->pos == NULL) {
						free(args);
						return;
					}
					args->ptr = (param *)calloc(1, sizeof(param));
					if (args->ptr == NULL) {
						free(args->pos);
						free(args);
						return;
					}
					memcpy(args->pos, pos, sizeof(smpl_set));
					memcpy(args->ptr, ptr, sizeof(param));
					strncpy(args->sampleset_dir, sampleset_dir, sizeof(args->sampleset_dir));

					// launch seperate thread for each instance of parameter
					pstats_launch_thread(args);
				}
			}
		}
	}
}

