#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdarg.h>
#include <stddef.h>
#include <setjmp.h>
#include <errno.h>
#include <cmocka.h>

#include <libubus.h>
#include <libubox/blobmsg_json.h>
#include <libubox/blobmsg.h>

#include <json-validator.h>
#include <json-c/json.h>
#include <json-editor.h>
#include <json-c/json_tokener.h>

#define UBUS_OUTPUT_FILE "/tmp/pstats_ubus_output.txt"

struct json_object *json_output = NULL;

/**
 *  group_setup function to setup the test modules
 *  @param state input parameter pointer to a pointer to test state
 */
static int group_setup(void **state)
{
	// Start ubusd if it is not running
	system("pidof ubusd || ubusd &");

	// Restart periodicstats
	system("kill $(pidof periodicstatsd) 2>/dev/null; ../../periodicstatsd &");

	return 0;
}

/**
 *  group_teardown function to cleanup the test modules
 *  @param state input parameter pointer to a pointer to test state
 */
static int group_teardown(void **state)
{
	// Stop periodicstatsd
	system("kill $(pidof periodicstatsd) 2>/dev/null");

	// Delete test logs
	unlink(UBUS_OUTPUT_FILE);

	return 0;
}

/**
 *  teardown function to cleanup the test modules
 *  @param state input parameter pointer to a pointer to test state
 */
static int teardown(void **state)
{
	if (json_output) {
		json_object_put(json_output); // Free the output JSON object
		json_output = NULL;
	}

	return 0;
}

/**
 *  get_ubus_call_output function to fill global json output variable with ubus call output
 *  @param ubus_cmd input parameter pointer to char string containg ubus call
 *  @param negative_case input parameter of type bool to identify negative scenarios
 */
static void get_ubus_call_output(const char *ubus_cmd, bool negative_case)
{
	char cmd[256];
	int fd = -1;
	char *str = NULL;

	// Put the output of ubus call into a string
	snprintf(cmd, sizeof(cmd), "%s > %s", ubus_cmd, UBUS_OUTPUT_FILE);
	printf("Execute Command : %s\n", cmd);
	int rc = system(cmd);
	assert_return_code(rc, errno);

	fd = open(UBUS_OUTPUT_FILE, O_RDONLY);
	assert_return_code(fd, errno);

	struct stat st;
	rc = fstat(fd, &st);


	if(negative_case) {
		assert_false(rc == 0 && st.st_size > 0);
		return;
	} else
		assert_true(rc == 0 && st.st_size > 0);

	str = calloc(1, (size_t)st.st_size + 1);
	printf("Command Output : %s\n", str);
	assert_non_null(str);

	ssize_t read_len = read(fd, str, (size_t)st.st_size);
	assert_int_equal((int)read_len, (int)st.st_size);

	// Parse the string to a json_object
	printf("JSON_OBJECT GET : %s\n", str);
	json_output = json_tokener_parse(str);

	if (fd >= 0)
		close(fd);
	if (str)
		free(str);
}

/**
 *  test_pstats_get_sampleset function to test configured sampleset
 *  @param state input parameter pointer to a pointer to test state
 */
static void test_pstats_get_sampleset(void **state)
{
	get_ubus_call_output("ubus call usp get '{\"path\":\"Device.PeriodicStatistics.\"}'", false);
}

/**
 *  test_pstats_get_sampleset_parameter function to test parameters of sampleset under uci configuration file.
 *  @param state input parameter pointer to a pointer to test state
 */
static void test_pstats_get_sampleset_parameter(void **state)
{
	get_ubus_call_output("ubus call usp get '{\"path\":\"Device.PeriodicStatistics.SampleSet.\"}'", false);
}

/**
 *   test_pstats_get_sampleset_parameter_values function to test actual statistics vales.
 *  @param state input parameter pointer to a pointer to test state
 */
static void test_pstats_get_sampleset_parameter_values(void **state)
{
	sleep(12);
	get_ubus_call_output("ubus call usp get '{\"path\":\"Device.PeriodicStatistics.SampleSet.1.Parameter.1.Values\"}'", false);
}

int main(void)
{
	const struct CMUnitTest tests[] = {
		cmocka_unit_test_teardown(test_pstats_get_sampleset, teardown),
		cmocka_unit_test_teardown(test_pstats_get_sampleset_parameter, teardown),
		cmocka_unit_test_teardown(test_pstats_get_sampleset_parameter_values, teardown),
	};

	return cmocka_run_group_tests(tests, group_setup, group_teardown);
}
